"""
Utility function for filtering frames based on statistics.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pointcloudset.config import OPS

if TYPE_CHECKING:
    from pointcloudset import PointCloud


def quantile_filter(
    pointcloud: PointCloud,
    dim: str,
    relation: str = ">=",
    cut_quantile: float = 0.5,
) -> PointCloud:
    """Filtering based on quantile values of dimension dim of the data.

    Args:
        pointcloud (PointCloud): PointCloud to be filtered.
        dim (str): Dimension to limit. Any column in data (not just x, y or z, but also
            "intensity")
        relation (str, optional): Any operator as string. Defaults to ">=".
        cut_quantile (float, optional): Quantile to compare to. Defaults to 0.5.

    Returns:
        PointCloud: PointCloud which fullfils the criteria.
    """
    cut_value = pointcloud.data[dim].quantile(cut_quantile)
    filter_array = OPS[relation](pointcloud.data[dim], cut_value)
    return pointcloud.apply_filter(filter_array.to_numpy())


def value_filter(
    pointcloud: PointCloud,
    dim: str,
    relation: str,
    value: float,
) -> PointCloud:
    """Limit the range of certain values in a PointCloud.

    Args:
        pointcloud (PointCloud): PointCloud to be filtered.
        dim (str): Dimension to limit. Any column in data (not just x, y or z, but also
            "intensity")
        relation (str): Any operator as string.
        value (float): Value to limit.

    Returns:
        PointCloud: PointCloud which fullfils the criteria.
    """

    bool_array = (OPS[relation](pointcloud.data[dim], value)).to_numpy()
    return pointcloud.apply_filter(bool_array)


def remove_radius_outlier(
    pointcloud: PointCloud, nb_points: int, radius: float
) -> PointCloud:
    """Function to remove points that have less than nb_points in a given
    sphere of a given radius.

    Args:
        pointcloud (PointCloud): PointCloud from which to remove points.
        nb_points (int): Number of points within the radius.
        radius (float): Radius of the sphere.

    Returns:
        PointCloud: PointCloud without outliers.
    """
    pcd = pointcloud.to_instance("open3d")
    _, index_to_keep = pcd.remove_radius_outlier(nb_points=nb_points, radius=radius)
    return pointcloud.apply_filter(index_to_keep)



"""
Functions to calculate differences and distances between entities.
"""

from pointcloudset.diff.nearest import calculate_distance_to_nearest
from pointcloudset.diff.origin import calculate_distance_to_origin
from pointcloudset.diff.plane import calculate_distance_to_plane
from pointcloudset.diff.point import calculate_distance_to_point
from pointcloudset.diff.pointcloud import calculate_distance_to_pointcloud

ALL_DIFFS = {
    "pointcloud": calculate_distance_to_pointcloud,
    "plane": calculate_distance_to_plane,
    "point": calculate_distance_to_point,
    "origin": calculate_distance_to_origin,
    "nearest": calculate_distance_to_nearest,
}



"""
Utility functions for geometric calculations with points.
"""

import numpy as np


def distance_to_point(point_a: np.ndarray, point_b: np.ndarray) -> float:
    """Calculate the Euclidean distance of a point to another point.

    Args:
        point_a (numpy.ndarray): [x, y, z]
        point_b (numpy.ndarray): [x, y, z]

    Returns:
        float: Euclidean distance between two points.

    Raises:
        ValueError: If any of the two points does not have 3 values.
    """
    if len(point_a) != 3:
        raise ValueError("point needs to have 3 values")
    if len(point_b) != 3:
        raise ValueError("point needs to have 3 values")
    return np.sqrt(
        (point_b[0] - point_a[0]) ** 2
        + (point_b[1] - point_a[1]) ** 2
        + (point_b[2] - point_a[2]) ** 2
    )



"""
Functions for plotting frames.
Used mainly by :func:`pointcloudset.pointcloud.PointCloud.plot` but could also be used
on its own.
"""
import math

import numpy as np
import plotly.express as px
import plotly.graph_objs as go


def plot_overlay(fig, pointcloud, overlay: dict, **kwargs):
    """Meta function to overlay the plot with other plots. Used in the main plot
    function.
    Not for standalone use.

    Args:
        fig (plotly.graph_objects.Figure): The original plot.
        pointcloud (PointCloud): The original pointcloud.
        overlay (dict): A dict with objects to overlay.
            For example overlay={"Cluster 1": cluster1, "plane1": plane_model}
            The name of the entry is used in tooltips and the value can either
            be a PointCloud or a model of a plane as numpy.ndarray [a,b,c,d] of a plane.
            Uses the plane equation a x + b y + c z + d = 0.

    Raises:
        ValueError: If the overlay value is wrong.

    Returns:
        plotly.graph_objects.Figure: Plot with all overlays.
    """
    fig.update_traces(opacity=0.7)
    fig.update_traces(
        marker=dict(size=1.5, line=dict(width=0)), selector=dict(mode="markers")
    )

    # Create color list dependend on length of overlay.
    if len(overlay) <= 9:
        colors = px.colors.qualitative.Plotly
    else:
        fact = len(overlay) / 24
        colors = px.colors.qualitative.Light24 * math.ceil(fact)

    for i, (name, from_dict) in enumerate(overlay.items()):
        marker_color = colors[i]
        if isinstance(from_dict, np.ndarray):
            fig = _plot_overlay_plane(
                fig,
                plane_model=from_dict,
                name=name,
                orig_pointcloud=pointcloud,
                colors=colors,
                i=i,
            )
        elif from_dict._has_data():
            fig = _plot_overlay_pointcloud(
                fig,
                pointcloud=from_dict,
                name=name,
                marker_color=marker_color,
                **kwargs,
            )
        else:
            raise ValueError(
                f"{from_dict} is not supported, use either a PointCloud or plane model"
            )

    return fig


def _plot_overlay_pointcloud(fig, pointcloud, name: str, marker_color: str, **kwargs):
    """Overlay the plot with another pointcloud.

    Args:
        fig (plotly.graph_objects.Figure): The original plot.
        pointcloud (PointCloud): The PointCloud to overlay.
        name (str): Name of the pointcloud to overlay, used for tooltips.
        marker_color (str): Color of the overlay.

    Returns:
        plotly.graph_objects.Figure: Plot with PointCloud overlayed with another
        PointCloud.
    """
    overlay_fig = pointcloud.plot(
        color=None, point_size=2.0, prepend_id=name + " ", opacity=0.7, **kwargs
    )
    overlay_fig.update_traces(marker_color=marker_color)
    overlay_trace = overlay_fig.data[0]
    return fig.add_trace(overlay_trace)


def _plot_overlay_plane(
    fig, plane_model: np.ndarray, name: str, orig_pointcloud, colors, i: int
):
    """Overlay the plot with plane.

    Args:
        fig (plotly.graph_objects.Figure): The original plot.
        plane_model (numpy.ndarray): [a,b,c,d], for the plane to overlay. Uses the plane
            equation a x + b y + c z + d = 0.
        name (str): Name of the plane to overlay, used for tooltips.
        orig_pointcloud:
        colors: colors from polot_overlay
        i (int): item number from plot_overlay

    Returns:
        plotly.graph_objects.Figure: Plot with PointCloud overlayed with plane.
    """
    bounding_box = orig_pointcloud.bounding_box
    x = np.linspace(bounding_box.x["min"], bounding_box.x["max"], 100)
    y = np.linspace(bounding_box.y["min"], bounding_box.y["max"], 100)
    z = np.linspace(bounding_box.z["min"], bounding_box.z["max"], 100)

    a, b, c, d = plane_model

    eps = 0.000001
    if (abs(c) < eps) & (abs(b) < eps):
        Y, Z = np.meshgrid(y, z)
        X = (-d - b * Y - c * Z) / a
    elif (abs(c) < eps) & (abs(b) > eps):
        X, Z = np.meshgrid(x, z)
        Y = (-d - a * X - c * Z) / b
    else:
        X, Y = np.meshgrid(x, y)
        Z = (-d - a * X - b * Y) / c

    colorscale = [[0, colors[i]], [1, colors[i]]]
    surfacecolor = np.ones(shape=X.shape)

    p2 = go.Surface(
        x=X,
        y=Y,
        z=Z,
        name=name,
        surfacecolor=surfacecolor,
        colorscale=colorscale,
        showscale=False,
        cmin=0,
        cmax=1,
        opacity=0.5,
    )

    fig.add_trace(p2)
    fig.update_layout(scene_aspectmode="data")
    return fig



"""
Utiliy functions for filtering of frames.
"""

from pointcloudset.filter.stat import (
    quantile_filter,
    remove_radius_outlier,
    value_filter,
)

ALL_FILTERS = {
    "QUANTILE": quantile_filter,
    "VALUE": value_filter,
    "RADIUSOUTLIER": remove_radius_outlier,
}



"""
Functions to calculate distances of points in pointcloud to origin.
"""

import numpy as np

from pointcloudset.diff.point import calculate_distance_to_point


def calculate_distance_to_origin(pointcloud, **kwargs):
    """Calculate the Euclidian distance to the origin (0,0,0) for each point in the pointcloud.

    Note:
        Adds the result as a new column to the data of the pointcloud.

    Args:
        pointcloud (PointCloud): PointCloud for which the Euclidean distance to the origin is
            calculated.

    Returns:
        PointCloud: PointCloud including Euclidean distances to the origin for each point.
    """
    point_a = np.array((0, 0, 0))

    return calculate_distance_to_point(pointcloud, target=point_a)



from __future__ import annotations

from typing import TYPE_CHECKING

import dask

if TYPE_CHECKING:
    from pointcloudset import PointCloud


def dataset_from_pointclouds(pointclouds: list[PointCloud]) -> dict:
    """Convert a list of pointcloud pointclouds to a new dataset.

    Args:
        pointclouds (list[PointCloud]): A list of pointclouds.

    Returns:
        dict: For convertion to dataset.
    """
    data = [dask.delayed(pointcloud.data) for pointcloud in pointclouds]
    timestamps = [pointcloud.timestamp for pointcloud in pointclouds]
    meta = {"orig_file": "from pointclouds list"}
    return {"data": data, "timestamps": timestamps, "meta": meta}



"""
Global definitions for the library.
"""

import operator

OPS = {
    ">": operator.gt,
    "<": operator.lt,
    ">=": operator.ge,
    "<=": operator.le,
    "==": operator.eq,
}

PLOTLYSIZELIMIT = 300000



from __future__ import annotations

from typing import TYPE_CHECKING

import open3d as o3d
from pyntcloud import PyntCloud

if TYPE_CHECKING:
    from pointcloudset import PointCloud


def from_open3d(open3d_data: o3d.geometry.PointCloud) -> dict:
    """Converts a open3d pointcloud to a PointCloud.

    Args:
        open3d_data (o3d.cpu.pybind.geometry.PointCloud): Open3D pointcloud which should be converted.

    Returns:
        dict: Pointcloud data.
    """
    pyntcloud_data = PyntCloud.from_instance("open3d", open3d_data)
    return {"data": pyntcloud_data.points}


def to_open3d(pointcloud: PointCloud) -> o3d.geometry.PointCloud:
    """Converts PointCloud to open3d PointCloud.

    Args:
        df (pd.DataFrame): Pointcloud dataframe with x,y,z,intensity.

    Returns:
        o3d.cpu.pybind.geometry.PointCloud: Open3d pointcloud object.
    """
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(pointcloud.points.xyz)
    return pcd



"""" Reading pointcloud 2.

`ROS <https://www.ros.org/>`_ bagfiles.

Parts of the code are from Willow Garage, Inc.

"""

# Software License Agreement (BSD License)
#
# Copyright (c) 2008, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided
# with the distribution.
# * Neither the name of Willow Garage, Inc. nor the names of its
# contributors may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import datetime
import math
import struct
import sys
from pathlib import Path
from typing import Generator, Literal, Union

import numpy as np
import pandas as pd
from dask import delayed
from rich.progress import track
from rosbags.rosbag1 import Reader as Reader1
from rosbags.rosbag2 import Reader as Reader2
from rosbags.serde import deserialize_cdr, ros1_to_cdr
from rosbags.typesys.types import sensor_msgs__msg__PointCloud2

_DATATYPES = {
    1: ("b", 1),
    2: ("B", 1),
    3: ("h", 2),
    4: ("H", 2),
    5: ("i", 4),
    6: ("I", 4),
    7: ("f", 4),
    8: ("d", 8),
}

PANDAS_TYPEMAPPING = {
    1: np.dtype("int8"),
    2: np.dtype("uint8"),
    3: np.dtype("int16"),
    4: np.dtype("uint16"),
    5: np.dtype("int32"),
    6: np.dtype("uint32"),
    7: np.dtype("float32"),
    8: np.dtype("float64"),
}


def dataset_from_ros(
    bagfile: Path,
    topic: str,
    start_frame_number: int = 0,
    end_frame_number: int = None,
    keep_zeros: bool = False,
    ext: Literal["BAG", "ROS2"] = "BAG",
) -> Union[dict, None]:
    """Reads a Dataset from a ROS1 bag of ROS2 mcap or db3 file.

    Args:
        bagfile (Path): Path to bag file.
        topic (str): `ROS <https://www.ros.org/>`_ topic that should be read
        start_frame_number (int, optional): Start pointcloud of pointcloud sequence to
            read. Defaults to 0.
        end_frame_number (int, optional): End pointcloud of pointcloud sequence to read.
            Defaults to None.
        keep_zeros (bool, optional): If ``True`` keep zeros in frames, if ``False``
            do not keep zeros in frames. Defaults to False.

    Returns:
        Union[dict, None]: Dict to generate Dataset.
    """

    data = []
    timestamps: list[datetime.datetime] = []
    meta = {"orig_file": bagfile.as_posix(), "topic": topic}

    if ext == "BAG":
        Reader = Reader1
        rosversion = 1
    elif ext == "ROS2":
        Reader = Reader2
        rosversion = 2
    else:
        raise ValueError(f"unexpected file extension got {ext}")
    with Reader(bagfile.as_posix()) as reader:
        connections = [x for x in reader.connections if x.topic == topic]

        frame = -1

        if not end_frame_number:
            end_frame_number = reader.topics[topic].msgcount

        for connection, timestamp, rawdata in track(
            reader.messages(connections=connections),
            total=end_frame_number - start_frame_number,
        ):
            frame = frame + 1
            if start_frame_number <= frame < end_frame_number:
                timestamp_datetime = datetime.datetime.fromtimestamp(timestamp * 1e-9)
                timestamps.append(timestamp_datetime)

                if rosversion == 1:
                    msg = deserialize_cdr(
                        ros1_to_cdr(rawdata, connection.msgtype), connection.msgtype
                    )
                elif rosversion == 2:
                    msg = deserialize_cdr(rawdata, connection.msgtype)
                data_of_frame = delayed(
                    _dataframe_from_message(msg, keep_zeros=keep_zeros)
                )
                data.append(data_of_frame)

    return {"data": data, "timestamps": timestamps, "meta": meta}


def _dataframe_from_message(
    message: sensor_msgs__msg__PointCloud2, keep_zeros: bool = False
) -> pd.DataFrame:
    columnnames = [field.name for field in message.fields]
    type_dict = {
        item.name: PANDAS_TYPEMAPPING[item.datatype] for item in message.fields
    }
    frame_raw = _read_points(message)
    frame_df = pd.DataFrame(np.array(list(frame_raw)), columns=columnnames)
    frame_df = frame_df.astype(type_dict)
    if not keep_zeros:
        frame_df = frame_df[
            (frame_df["x"] != 0.0) & (frame_df["y"] != 0.0) & (frame_df["z"] != 0.0)
        ]
        frame_df["original_id"] = frame_df.index
        frame_df = frame_df.astype({"original_id": "uint32"})
        frame_df = frame_df.reset_index(drop=True)
    return frame_df


def _get_struct_fmt(is_bigendian, fields, field_names=None):
    """
    code from from Willow Garage, Inc.
    """
    fmt = ">" if is_bigendian else "<"

    offset = 0
    for field in (
        f
        for f in sorted(fields, key=lambda f: f.offset)
        if field_names is None or f.name in field_names
    ):
        if offset < field.offset:
            fmt += "x" * (field.offset - offset)
            offset = field.offset
        if field.datatype not in _DATATYPES:
            print(
                f"Skipping unknown PointField datatype {field.datatype}",
                file=sys.stderr,
            )
        else:
            datatype_fmt, datatype_length = _DATATYPES[field.datatype]
            fmt += field.count * datatype_fmt
            offset += field.count * datatype_length

    return fmt


def _read_points(
    cloud: sensor_msgs__msg__PointCloud2, field_names=None, skip_nans=False, uvs=[]
) -> Generator:
    """
    Read points from a PointCloud2 message.
    code from from Willow Garage, Inc.

    @param cloud: The point cloud to read from.
    @type  cloud: L{sensor_msgs.PointCloud2}
    @param field_names: The names of fields to read. If None, read all fields. [default: None]
    @type  field_names: iterable
    @param skip_nans: If True, then don't return any point with a NaN value.
    @type  skip_nans: bool [default: False]
    @param uvs: If specified, then only return the points at the given coordinates. [default: empty list]
    @type  uvs: iterable
    @return: Generator which yields a list of values for each point.
    @rtype:  generator
    """
    assert isinstance(
        cloud, sensor_msgs__msg__PointCloud2
    ), "cloud is not a PointCloud2"
    fmt = _get_struct_fmt(cloud.is_bigendian, cloud.fields, field_names)
    width, height, point_step, row_step, data, isnan = (
        cloud.width,
        cloud.height,
        cloud.point_step,
        cloud.row_step,
        cloud.data,
        math.isnan,
    )
    unpack_from = struct.Struct(fmt).unpack_from

    if skip_nans:
        if uvs:
            for u, v in uvs:
                p = unpack_from(data, (row_step * v) + (point_step * u))
                has_nan = False
                for pv in p:
                    if isnan(pv):
                        has_nan = True
                        break
                if not has_nan:
                    yield p
        else:
            for v in range(height):
                offset = row_step * v
                for u in range(width):
                    p = unpack_from(data, offset)
                    has_nan = False
                    for pv in p:
                        if isnan(pv):
                            has_nan = True
                            break
                    if not has_nan:
                        yield p
                    offset += point_step
    else:
        if uvs:
            for u, v in uvs:
                yield unpack_from(data, (row_step * v) + (point_step * u))
        else:
            for v in range(height):
                offset = row_step * v
                for u in range(width):
                    yield unpack_from(data, offset)
                    offset += point_step



from __future__ import annotations

import datetime
import warnings
from pathlib import Path
from typing import Literal, Union

import numpy as np
import open3d
import pandas
import plotly
import plotly.express as px
import pyntcloud

from pointcloudset.config import PLOTLYSIZELIMIT
from pointcloudset.diff import ALL_DIFFS
from pointcloudset.filter import ALL_FILTERS
from pointcloudset.io import (
    POINTCLOUD_FROM_FILE,
    POINTCLOUD_FROM_INSTANCE,
    POINTCLOUD_TO_FILE,
    POINTCLOUD_TO_INSTANCE,
)
from pointcloudset.plot.pointcloud import plot_overlay
from pointcloudset.pointcloud_core import PointCloudCore


class PointCloud(PointCloudCore):
    """
    PointCloud Class with one pointcloud of lidar measurements, laser scanning,
    photogrammetry  or simular.

    One PointCloud consists mainly of
    `PyntCloud <https://pyntcloud.readthedocs.io/en/latest/>`_
    pointcloud
    (`PyntCloud.points <https://pyntcloud.readthedocs.io/en/latest/points.html#points>`_)
    and a pandas.DataFrame (.data) with all the associated data.

    Note that the index of the points is not preserved when applying processing. This
    is necessary since `PyntCloud <https://pyntcloud.readthedocs.io/en/latest/>`_
    does not allow to pass the index. Therefore, a new PointCloud object is generated at
    each processing stage.

    Developer notes:
        * All operations have to act on both, pointcloud, data and keep the timestamp.
        * All processing methods need to return another PointCloud.

    Examples:

        .. code-block:: python

            testbag = Path().cwd().parent.joinpath("tests/testdata/test.bag")
            testset = pointcloudset.Dataset(testbag,topic="/os1_cloud_node/points",
                keep_zeros=False)
            testpointcloud = testset[0]
    """

    @classmethod
    def from_file(
        cls,
        file_path: Path,
        timestamp: str | datetime.datetime = "from_file",
        **kwargs,
    ):
        """Extract data from file and construct a PointCloud with it. Uses
        `PyntCloud <https://pyntcloud.readthedocs.io/en/latest/>`_ as
        backend.

        Args:
            file_path (pathlib.Path): Path of file to read.
            use_file_timestamp (bool): use the file creation date as timestamp.
                Defaults to True.
            timestamp (str | datetime.datetime): timestamp of pointcloud. If "from_file"
                then the timesamp is taken from file creation datetimne.
                (Defaults to "from_file")
            **kwargs: Keyword arguments to pass to func.

        Returns:
            PointCloud: PointCloud with timestamp last modified.

        Raises:
            ValueError: If file format is not supported.
            TypeError: If file_path is no Path object.
        """
        if not isinstance(file_path, Path):
            raise TypeError("Expecting a Path object for file_path")
        ext = file_path.suffix[1:].upper()
        if ext not in POINTCLOUD_FROM_FILE:
            raise ValueError(
                "Unsupported file format; supported formats are: {}".format(
                    list(POINTCLOUD_FROM_FILE)
                )
            )
        file_path_str = file_path.as_posix()
        if timestamp == "from_file":
            timestamp = datetime.datetime.utcfromtimestamp(file_path.stat().st_mtime)
        pyntcloud_in = pyntcloud.PyntCloud.from_file(file_path_str, **kwargs)
        return cls(
            data=pyntcloud_in.points, orig_file=file_path_str, timestamp=timestamp
        )

    def to_file(self, file_path: Path = Path(), **kwargs) -> None:
        """Exports the pointcloud as to a file for use with
        `CloudCompare <https://www.danielgm.net/cc/ake>`_ or similar tools.
        Currently not all attributes of a pointcloud are saved so some information
        is lost when using this function.
        Uses `PyntCloud <https://pyntcloud.readthedocs.io/en/latest/>`_ as
        backend.

        Args:
            file_path (pathlib.Path, optional): Destination. Defaults to the folder of
                the bag file and csv with the timestamp of the pointcloud.
            **kwargs: Keyword arguments to pass to func.

        Raises:
            ValueError: If file format is not supported.
        """
        ext = file_path.suffix[1:].upper()
        if ext not in POINTCLOUD_TO_FILE:
            raise ValueError(
                "Unsupported file format; supported formats are: {}".format(
                    list(POINTCLOUD_TO_FILE)
                )
            )

        orig_file_name = Path(self.orig_file).stem
        if file_path == Path():
            # defaulting to csv file
            filename = f"{orig_file_name}_timestamp_{self.timestamp}.csv"
            destination_folder = Path(self.orig_file).parent.joinpath(filename)
        else:
            destination_folder = file_path

        kwargs["file_path"] = destination_folder
        kwargs["pointcloud"] = self

        POINTCLOUD_TO_FILE[ext](**kwargs)

    @classmethod
    def from_instance(
        cls,
        library: Literal[
            "PANDAS",
            "PYNTCLOUD",
            "OPEN3D",
            "DATAFRAME",
        ] = "PANDAS",
        instance: (
            pandas.DataFrame | pyntcloud.PyntCloud | open3d.geometry.PointCloud
        ) = pandas.DataFrame(),
        **kwargs,
    ) -> PointCloud:
        """Converts a library instance to a pointcloudset PointCloud.

        Args:
            library (str): Name of the library.\n
                If PYNTCLOUD: :func:`pointcloudset.io.pointcloud.pyntcloud.from_pyntcloud`\n
                If OPEN3D: :func:`pointcloudset.io.pointcloud.open3d.from_open3d`\n
                If DATAFRAME: :func:`pointcloudset.io.pointcloud.pandas.from_dataframe`\n
                If PANDAS: :func:`pointcloudset.io.pointcloud.pandas.from_dataframe`
            instance
                (Union[pandas.DataFrame, pyntcloud.PyntCloud, open3d.geometry.PointCloud]):
                Library instance to convert.
            **kwargs: Keyword arguments to pass to func.

        Returns:
            PointCloud: Derived from the instance.

        Raises:
            ValueError: If instance is not supported.

        Examples:

            .. code-block:: python

                testpointcloud = from_instance("OPEN3D", open3d_pointcloud)

        """
        library = library.upper()
        if library not in POINTCLOUD_FROM_INSTANCE:
            raise ValueError(
                "Unsupported library; supported libraries are: {}".format(
                    list(POINTCLOUD_FROM_INSTANCE)
                )
            )
        else:
            return cls(**POINTCLOUD_FROM_INSTANCE[library](instance, **kwargs))

    def to_instance(
        self, library: Literal["PYNTCLOUD", "OPEN3D", "DATAFRAME", "PANDAS"], **kwargs
    ) -> (
        pyntcloud.PyntCloud
        | open3d.geometry.PointCloud
        | pandas.DataFrame
        | pandas.DataFrame
    ):
        """Convert PointCloud to another library instance.

        Args:
            library (str): Name of the library.\n
                If PYNTCLOUD: :func:`pointcloudset.io.pointcloud.pyntcloud.to_pyntcloud`\n
                If OPEN3D: :func:`pointcloudset.io.pointcloud.open3d.to_open3d`\n
                If DATAFRAME: :func:`pointcloudset.io.pointcloud.pandas.to_dataframe`\n
                If PANDAS: :func:`pointcloudset.io.pointcloud.pandas.to_dataframe`
            **kwargs: Keyword arguments to pass to func.

        Returns:
            Union[ pandas.DataFrame, pyntcloud.PyntCloud, open3d.geometry.PointCloud ]:
            The derived instance.

        Raises:
            ValueError: If library is not suppored.

        Examples:

            .. code-block:: python

                open3d_pointcloud = testpointcloud.to_instance("OPEN3D")
        """
        library = library.upper()
        if library not in POINTCLOUD_TO_INSTANCE:
            raise ValueError(
                "Unsupported library; supported libraries are: {}".format(
                    list(POINTCLOUD_TO_INSTANCE)
                )
            )

        return POINTCLOUD_TO_INSTANCE[library](self, **kwargs)

    def plot(
        self,
        color: None | str = None,
        overlay: dict | None = None,
        point_size: float = 2,
        prepend_id: str = "",
        hover_data: Union(list[str], bool) = None,
        **kwargs,
    ) -> plotly.graph_objs.Figure:
        """Plot a PointCloud as a 3D scatter plot with `Plotly <https://plotly.com/>`_.
        It handles plots of single pointclouds and overlay with other objects, such as
        other pointclouds from clustering or planes from plane segmentation.

        You can also pass arguments to the `Plotly <https://plotly.com/>`_
        express function :func:`plotly.express.scatter_3d`.

        Args:
            pointcloud (PointCloud): The pointcloud to plot.
            color (str or None): Which column to plot. For example "intensity".
                Defaults to None.
            overlay (dict, optional): Dict with PointClouds to overlay.
                {"Cluster 1": cluster1,"Plane 1": plane_model}\n
                See also: :func:`pointcloudset.plot.pointcloud.plot_overlay`\n
                Defaults to empty.
            point_size (float, optional): Size of each point. Defaults to 2.
            prepend_id (str, optional): String before point id to display in hover.
                Defaults to empty.
            hover_data (list(str) or True, optional): Data columns to display in hover. If True
                then all the columns are are show in the hover.
                Defaults to None.
            **kwargs: Keyword arguments to pass to func.

        Returns:
            plotly.graph_objs.Figure: The interactive Plotly plot, best used inside a
            Jupyter Notebook.

        Raises:
            ValueError: If the color column name is not in the data.
        """
        if color is not None and color not in self.data.columns:
            raise ValueError(f"choose any of {list(self.data.columns)} or None")

        if len(self) > PLOTLYSIZELIMIT:
            warnings.warn(
                f"""Pointcloud above limit of {PLOTLYSIZELIMIT}.
                Plotting might fail or take a long time.
                Consider donwsampling before plotting.
                for example: pointcloud.random_down_sample(10000).plot()"""
            )

        ids = [prepend_id + "id=" + str(i) for i in range(self.data.shape[0])]

        show_hover = True
        if hover_data is None:
            show_hover = False
        elif hover_data:
            hover_data = list(self.data.columns)
        elif isinstance(hover_data, list) & len(hover_data) > 0:
            if self.has_original_id:
                default = ["original_id"]
                hover_data = hover_data + default
            hover_data = list(set(hover_data))
            if any(x not in self.data.columns for x in hover_data):
                raise ValueError(f"choose a list of {list(self.data.columns)} or []")

        fig = px.scatter_3d(
            self.data,
            x="x",
            y="y",
            z="z",
            color=color,
            hover_name=ids,
            hover_data=hover_data,
            title=self.timestamp_str,
            **kwargs,
        )

        if overlay:
            fig = plot_overlay(
                fig,
                self,
                overlay,
                hover_data=hover_data,
                **kwargs,
            )
        fig.update_layout(scene_aspectmode="data")
        if not show_hover:
            fig.update_layout(hovermode=False)

        fig.update_traces(
            marker=dict(size=point_size, line=dict(width=0)),
            selector=dict(mode="markers"),
        )

        return fig

    def diff(
        self,
        name: Literal["origin", "plane", "pointcloud", "point", "nearest"],
        target: None | PointCloud | np.ndarray = None,
        **kwargs,
    ) -> PointCloud:
        """Calculate differences and distances to the origin, plane, point and pointcloud.

        Args:
            name (str):
                "origin": :func:`pointcloudset.diff.origin.calculate_distance_to_origin` \n
                "plane": :func:`pointcloudset.diff.plane.calculate_distance_to_plane` \n
                "pointcloud": :func:`pointcloudset.diff.pointcloud.calculate_distance_to_pointcloud` \n
                "point": :func:`pointcloudset.diff.point.calculate_distance_to_point` \n
                "nearest": :func:`pointcloudset.diff.point.calculate_distance_to_nearest` \n
            target (Union[None, PointCloud, numpy.ndarray], optional): Pass argument
                according to chosen object. Defaults to None.
            **kwargs: Keyword arguments to pass to func.

        Returns:
            PointCloud: New PointCloud with added column of the differences.

        Raises:
            ValueError: If name is not supported.

        Examples:

            .. code-block:: python

                newpointcloud = testpointcloud.diff("pointcloud", targetpointcloud)
        """
        if name in ALL_DIFFS:
            ALL_DIFFS[name](pointcloud=self, target=target, **kwargs)
            return self
        else:
            raise ValueError("Unsupported diff. Check docstring")

    def filter(
        self, name: Literal["quantile", "value", "radiusoutlier"], *args, **kwargs
    ) -> PointCloud:
        """Filters a PointCloud according to criteria.

        Args:
            name (str):
                "quantile": :func:`pointcloudset.filter.stat.quantile_filter` \n
                "value": :func:`pointcloudset.filter.stat.value_filter` \n
                "radiusoutlier": :func:`pointcloudset.filter.stat.remove_radius_outlier` \n
            *args: Positional arguments to pass to func.
            **kwargs: Keyword arguments to pass to func.

        Returns:
            PointCloud: PointCloud which fullfils the criteria.

        Raises:
            ValueError: If name is not supported.

        Examples:

            .. code-block:: python

                filteredpointcloud = testpointcloud.filter("quantile","intensity","==",0.5)

            .. code-block:: python

                filteredpointcloud = testpointcloud.filter("value","intensity",">",100)
        """
        name = name.upper()
        if name in ALL_FILTERS:
            return ALL_FILTERS[name](self, *args, **kwargs)
        else:
            raise ValueError("Unsupported filter. Check docstring")

    def limit(self, dim: str, minvalue: float, maxvalue: float) -> PointCloud:
        """Limit the range of certain values in pointcloudset PointCloud. Can be chained together.

        Args:
            dim (str): Dimension to limit, any column in data not just x, y, or z.
            minvalue (float): Min value to limit. (greater equal)
            maxvalue (float): Max value to limit. (smaller equal)

        Returns:
            PointCloud: Limited pointcloud, where rows which did not match the criteria were
            dropped.

        Examples:

            .. code-block:: python

                limitedpointcloud = testpointcloud.limit("x", -1.0, 1.0).limit("intensity", 0.0, 50.0)
        """
        if maxvalue < minvalue:
            raise ValueError("maxvalue must be greater than minvalue")
        return self.filter("value", dim, ">=", minvalue).filter(
            "value", dim, "<=", maxvalue
        )

    def limit_less(self, dim: str, value: float) -> PointCloud:
        """Limit the range if a diminsion to a value.
            Same as filter("value", dim, "<", value)

        Args:
            dim (str): Dimension to limit, any column in data not just x, y, or z.
            value (float): Min value to limit. (less)

        Returns:
            PointCloud: Limited pointcloud, where rows which did not match the criteria
            were dropped.

        Examples:

            .. code-block:: python

                limitedpointcloud = testpointcloud.limit_less("x",1.0)
        """
        return self.filter("value", dim, "<", value)

    def limit_greater(self, dim: str, value: float) -> PointCloud:
        """Limit the range if a diminsion to a value.
            Same as filter("value", dim, ">", value)

        Args:
            dim (str): Dimension to limit, any column in data not just x, y, or z.
            value (float): Value to limit. (greater)

        Returns:
            PointCloud: Limited pointcloud, where rows which did not match the criteria
            were dropped.

        Examples:

            .. code-block:: python

                limitedpointcloud = testpointcloud.limit_greater("x",10.0)
        """
        return self.filter("value", dim, ">", value)

    def apply_filter(self, filter_result: np.ndarray | list[int]) -> PointCloud:
        """Generating a new PointCloud by removing points according to a call of the
        filter method.

        Args:
            filter_result (Union[numpy.ndarray, list[int]]): Filter result.

        Returns:
            PointCloud: PointCloud with filtered rows and reindexed data and points.

        Raises:
            TypeError: If the filter_result has the wrong type.

        """
        if isinstance(filter_result, np.ndarray):
            # dataframe and pyntcloud based filters
            new_data = self.data.loc[filter_result].reset_index(drop=True)
        elif isinstance(filter_result, list):
            # from open3d filters
            new_data = self.data.iloc[filter_result].reset_index(drop=True)
        else:
            raise TypeError(
                "Wrong filter_result expecting array with boolean values or"
                "list of indices"
            )
        return PointCloud(new_data, timestamp=self.timestamp)

    def get_cluster(self, eps: float, min_points: int) -> pandas.DataFrame:
        """Get the clusters based on
        :meth:`open3d:open3d.geometry.PointCloud.cluster_dbscan`.
        Process further with :func:`pointcloudset.pointcloud.PointCloud.take_cluster`.

        Args:
            eps (float): Density parameter that is used to find neighboring points.
            min_points (int): Minimum number of points to form a cluster.

        Returns:
            pandas.DataFrame: Dataframe with list of clusters.
        """
        labels = np.array(
            self.to_instance("open3d").cluster_dbscan(
                eps=eps, min_points=min_points, print_progress=False
            )
        )
        return pandas.DataFrame(labels, columns=["cluster"])

    def take_cluster(
        self, cluster_number: int, cluster_labels: pandas.DataFrame
    ) -> PointCloud:
        """Takes only the points belonging to the cluster_number.

        Args:
            cluster_number (int): Cluster ID to keep.
            cluster_labels (pandas.DataFrame): Clusters generated with
                :func:`pointcloudset.pointcloud.PointCloud.get_cluster`.

        Returns:
            PointCloud: PointCloud with selected cluster.
        """
        bool_array = (cluster_labels["cluster"] == cluster_number).values
        return self.apply_filter(bool_array)

    def plane_segmentation(
        self,
        distance_threshold: float,
        ransac_n: int,
        num_iterations: int,
        return_plane_model: bool = False,
    ) -> PointCloud | dict:
        """Segments a plane in the point cloud using the RANSAC algorithm.
        Based on :meth:`open3d:open3d.geometry.PointCloud.segment_plane`.

        Args:
            distance_threshold (float): Max distance a point can be from the plane
                model, and still be considered as an inlier.
            ransac_n (int):  Number of initial points to be considered inliers in
                each iteration.
            num_iterations (int): Number of iterations.
            return_plane_model (bool, optional): Return also plane model parameters
                if ``True``. Defaults to ``False``.

        Returns:
            PointCloud or dict: PointCloud with inliers or a dict of PointCloud with inliers and the
            plane parameters.
        """
        pcd = self.to_instance("open3d")
        plane_model, inliers = pcd.segment_plane(
            distance_threshold=distance_threshold,
            ransac_n=ransac_n,
            num_iterations=num_iterations,
        )
        if len(self) > 200:
            warnings.warn(
                """Might not produce reproduceable resuts, if the number of points
                is high. Try to reduce the area of interest before using
                plane_segmentation. Caused by open3D."""
            )
        inlier_pointcloud = self.apply_filter(inliers)
        if return_plane_model:
            return {"PointCloud": inlier_pointcloud, "plane_model": plane_model}
        else:
            return inlier_pointcloud

    def random_down_sample(self, number_of_points: int) -> PointCloud:
        """Function to downsample input pointcloud into output pointcloud randomly.
        Made

        Args:
            number_of_points ([int]): number_of_points

        Returns:
            PointCloud: subsampled PointCloud
        """
        new_data = self.data.sample(number_of_points).reset_index()
        return PointCloud(new_data, timestamp=self.timestamp)

    def _add_original_id_from_index(self) -> PointCloud:
        """Add orginal ID column from index."""
        return self._add_column("original_id", self.data.index)



from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from pointcloudset import PointCloud


def from_dataframe(df: pd.DataFrame) -> dict:
    """Converts pandas DataFrame to a PointCloud.

    Args:
        df (pandas.DataFrame): Pandas DataFrame to convert.

    Returns:
        dict: Returns data for PointCloud.
    """
    return {"data": df}


def to_dataframe(pointcloud: PointCloud) -> pd.DataFrame:
    """Converts a PointCloud to a pandas DataFrame.

    Args:
        pointcloud (PointCloud): PointCloud to convert.

    Returns:
        pandas.DataFrame: Data of PointCloud.
    """
    return pointcloud.data



from __future__ import annotations

import datetime

import dask
import pandas as pd
from dask.delayed import Delayed, DelayedLeaf


class DatasetCore:
    """
    DatasetCore Class with all the main methods and properties of the Dataset Class.
    """

    def __init__(
        self,
        data: list[dask.delayed.DelayedLeaf] = [],
        timestamps: list[datetime.datetime] = [],
        meta: dict = {"orig_file": "", "topic": ""},
    ) -> None:
        self.data = data
        self.timestamps = timestamps
        self.meta = meta
        self._check()

    @property
    def start_time(self) -> datetime.datetime:
        """
        Returns:
            datetime.datetime: Time of first pointcloud in Dataset.
        """
        return self.timestamps[0]

    @property
    def end_time(self) -> datetime.datetime:
        """
        Returns:
            datetime.datetime: Time of last pointcloud in Dataset.
        """
        return self.timestamps[-1]

    @property
    def duration(self) -> datetime.timedelta:
        """
        Returns:
            datetime.timedelta: Duration between first and last
            pointcloud in Dataset.
        """
        return self.timestamps[-1] - self.timestamps[0]

    @property
    def daskdataframe(self) -> dask.dataframe.core.DataFrame:
        """
        Returns:
            dask.dataframe.DataFrame: Dask DataFrame with data of Dataset.
        """
        return dask.dataframe.from_delayed(self.data)

    @property
    def bounding_box(self) -> pd.DataFrame:
        """The axis aligned boundary box of the whole dataset as a :class:`pandas.DataFrame`."""

        def bb(pc):
            return pc.bounding_box

        list_of_bb = self.apply(bb, warn=False).compute()
        bb_all_df = pd.concat(list_of_bb)
        return pd.DataFrame([bb_all_df.min(), bb_all_df.max()], index=["min", "max"])

    def __len__(self) -> int:
        """Number of available frames (i.e. Lidar messages)"""
        return len(self.data)

    def __str__(self):
        return f"Lidar Dataset with {len(self)} pointcloud(s)"

    def __repr__(self) -> str:
        return (
            f"""{self.__class__.__name__}({self.data},{self.timestamps},{self.meta})"""
        )

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n >= len(self):
            raise StopIteration

        result = self[self.n]
        self.n += 1
        return result

    def _agg(self, agg: str | list | dict) -> dask.dataframe.DataFrame:
        """Aggregate using one or more operations over the whole dataset.
            Similar to pandas agg. Used dask dataframes with parallel processing.

        Example:
            .. code-block:: python

                dataset.agg("max")
                datset.agg(["min","max","mean","std"])
                datset.agg({"x" : ["min","max","mean","std"]})

        Args:
            agg (Union[str, list, dict]): [description]

        Returns:
            dask.dataframe.DataFrame: use .compute to get final result.
        """
        if self.has_original_id:
            data = self.daskdataframe.groupby("original_id").agg(agg)
            data["N"] = self.daskdataframe.groupby("original_id").size()
            data["original_id"] = data.index
            data = data.reset_index(drop=True)
        else:
            data = self.daskdataframe
            data["dummy"] = data.index
            data = data.groupby("dummy").agg(agg)
        return data

    def has_pointclouds(self) -> bool:
        """Check if Dataset has PointCloud.

        Returns:
            bool: ``True`` if the Dataset does contain PointClouds, ``False`` if Dataset
            does not contain any PointClouds.
        """
        return len(self) > 0

    def get_pointclouds_between_timestamps(
        self, start_time: datetime.datetime, end_time: datetime.datetime
    ) -> DatasetCore:
        """Select pointclouds between start_time and end_time.

        Args:
            start_time (datetime.datetime): Timestamp of first pointcloud.
            end_time (datetime.datetime): Timestamp of last pointcloud.

        Returns:
            Dataset: Dataset with PointClouds between two timestamps.

        Raises:
            ValueError: If start_time is bigger than end_time.
        """
        if start_time >= end_time:
            raise ValueError("start_time must be smaller than end_time")
        start_i = self._get_pointcloud_number_from_time(start_time)
        end_i = self._get_pointcloud_number_from_time(end_time) + 1
        return self[start_i:end_i]

    def _get_pointcloud_number_from_time(self, time: datetime.datetime) -> int:
        """Get the pointcloud number from a timestamp.

        Args:
            time (datetime.datetime): The time of interest.

        Returns:
            int: PointCloud number.

        Raises:
            ValueError: If time is outside of range.
        """
        if time < self.start_time or time > self.end_time:
            raise ValueError("time is outside of range")
        return min(
            range(len(self.timestamps)), key=lambda i: abs(self.timestamps[i] - time)
        )

    def _check(self):
        assert "orig_file" in self.meta, "meta data does not contain orig_file"
        if len(self) > 0:
            assert len(self.timestamps) == len(
                self.data
            ), f"Length of timestamps {len(self.timestamps)} do not match the data {len(self.data)}"

            if not pd.Series(self.timestamps).is_monotonic_increasing:
                raise ValueError("Timestamps are not monotonic increasing")

            assert isinstance(
                self.data[0], (DelayedLeaf, Delayed)
            ), f"data needs to be a dask delayed object got {type(self.data[0])}"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pointcloudset import PointCloud


def write_csv(pointcloud: PointCloud, file_path: Path) -> None:
    """Exports the pointcloud as a csv for use with cloud compare or similar tools.
    Currently not all attributes of a pointcloud are saved so some information is lost when
    using this function.

    Args:
        file_path (Path): Destination.
    """
    pointcloud.data.to_csv(file_path, index=False)



"""
Functions to calculate differences between the pointcloud nearest points in another one.
"""
import numpy as np


def calculate_distance_to_nearest(pointcloud, target):
    """Calculate the distance for each point in a pointcloud to the nearest points in
    the target pointcloud.

    Note:
        Adds the results to the data of the pointcloud.

    Args:
        pointcloud (PointCloud): PointCloud for which the differences to the target are
            calculated.
        target (PointCloud): PointCloud to calcluate the distances to.

    Returns:
        PointCloud: PointCloud with the colum "distance to nearest point".

    Raises:
        ValueError: If distance ot nearest points already exits.

    """
    if "distance to nearest point" in pointcloud.data.columns:
        raise ValueError("distance to nearest point already exists.")
    pc1 = pointcloud.to_instance("open3d")
    pc2 = target.to_instance("open3d")
    distance = pc1.compute_point_cloud_distance(pc2)
    pointcloud._add_column("distance to nearest point", np.array(distance))
    return pointcloud



from __future__ import annotations

from pathlib import Path
from typing import Union

import click  # needed for documentation
import typer
from pyntcloud.io import TO_FILE
from rich.console import Console
from rosbags.highlevel import AnyReader

import pointcloudset
from pointcloudset import Dataset

app = typer.Typer()
console = Console()

TO_FILE_PYNTCLOUD = list(TO_FILE.keys())
TO_FILE_CLI = TO_FILE_PYNTCLOUD.append("POINTCLOUDSET")


@app.command()
def convert(
    ros_file: str,
    topic: str = typer.Option("/os1_cloud_node/points", "--topic", "-t"),
    folder_to_write: str = typer.Option(".", "--output-dir", "-d"),
    output_format: str = typer.Option("POINTCLOUDSET", "--output-format", "-o"),
    start_frame_number: int = typer.Option(0, "--start", "-s"),
    end_frame_number: Union[int, None] = typer.Option(None, "--end", "-e"),
    keep_zeros: bool = False,
):
    """The main CLI function to convert ROS1 and ROS2 files to pointcloudset or files
    supported by pyntloud.

    Examples:

    convert all ROS1 bag files in a directory
    $ pointcloudset convert -d converted .

    convert all frames of bagfile xyz.bag into csv files
    $ pointcloudset convert -o csv -d converted_csv xyz.bag

    convert a ROS2 directoy to a pointcloudset file
    $ pointcloudset convert -d converted something_ros2

    convert the first 10 frames of a bag file int0las files
    $ pointcloudset convert -o las -d converted_las --start 1 --end 10 xyz.bag
    """
    console.line()
    console.rule(f"pointcloudset {pointcloudset.__version__}")
    bagfile_paths = _gen_file_paths(ros_file)
    console.rule(output_format)
    with console.status("Converting...", spinner="runner"):
        for bagfile_path in bagfile_paths:
            console.rule(f"converting {bagfile_path.name} ...", style="blue")

            folder_to_write_path = _gen_folder(
                folder_to_write, bagfile_path, output_format
            )

            if output_format == "POINTCLOUDSET":
                _convert_one_bag2dir(
                    ros_file=bagfile_path,
                    topic=topic,
                    start_frame_number=start_frame_number,
                    end_frame_number=end_frame_number,
                    keep_zeros=keep_zeros,
                    folder_to_write=folder_to_write_path,
                )
                console.print(
                    f"{Path(bagfile_path).name} converted to {folder_to_write_path}"
                )
            elif output_format.upper() in TO_FILE_PYNTCLOUD:
                _convert_bag2files(
                    topic,
                    start_frame_number,
                    end_frame_number,
                    output_format,
                    bagfile_path,
                    folder_to_write_path,
                )

            else:
                raise typer.BadParameter(f"only one of {TO_FILE_CLI} is allowed")
    console.rule("Done :sake:")


@app.command()
def topics(ros_file: str):
    """List all ROS PointCloud2 topics which can be converted.

    Args:
        ros_file (str): ROS1 or ROS2 file or directory

    Examples:

    show all PointCloud2 topis in a  ROS1 bag file
    $ pointcloudset topics test.bag
    """
    with AnyReader([Path(ros_file)]) as reader:
        all_topics = reader.topics
    pointcloud_topics = {
        k: f"{v.msgtype} with {v.msgcount} messages"
        for k, v in all_topics.items()
        if v.msgtype == "sensor_msgs/msg/PointCloud2"
    }

    if len(pointcloud_topics) == 0:
        console.rule("no pointcloud topics found")
    else:
        console.rule(f"found {len(pointcloud_topics)} PointCloud2 topics :sake: ")
        console.print(pointcloud_topics)
    console.rule()


def _convert_one_bag2dir(
    ros_file: Path,
    topic: str,
    start_frame_number: int = 0,
    end_frame_number: int = None,
    keep_zeros: bool = False,
    folder_to_write: Path = Path(),
):
    if not ros_file.exists():
        raise typer.BadParameter(f"{ros_file} does not exist")
    dataset = Dataset.from_file(
        file_path=ros_file,
        topic=topic,
        start_frame_number=start_frame_number,
        end_frame_number=end_frame_number,
        keep_zeros=keep_zeros,
    )
    if len(dataset) > 0:
        dataset.to_file(
            file_path=folder_to_write,
            use_orig_filename=False,
        )
    else:
        console.print("no data, skipping")


def _gen_file_paths(file_name):
    if file_name == ".":
        bagfile_paths = list(Path.cwd().glob("*.bag"))
    else:
        bagfile_paths = [Path(file_name)]
    return bagfile_paths


def _gen_folder(folder_to_write: str, ros_file_path: str, output_format: str) -> Path:
    """Generate the folder to write the converted files to."""
    suffix = "_pointcloudset" if output_format == "POINTCLOUDSET" else ""
    folder_to_write_path = Path(folder_to_write).joinpath(
        Path(ros_file_path).stem + suffix
    )
    folder_to_write_path.mkdir(exist_ok=False, parents=True)
    return folder_to_write_path


def _convert_bag2files(
    topic,
    start_frame_number,
    end_frame_number,
    output_format,
    ros_file_path,
    folder_to_write_path,
):
    """Converting a bagfile to files for each frame. Using pyntcloud

    Args:
        topic (_type_): _description_
        start_frame_number (_type_): _description_
        end_frame_number (_type_): _description_
        output_format (_type_): _description_
        ros_file_path (_type_): _description_
        folder_to_write_path (_type_): _description_
    """
    dataset = Dataset.from_file(
        file_path=ros_file_path,
        topic=topic,
        keep_zeros=False,
        start_frame_number=start_frame_number,
        end_frame_number=end_frame_number,
    )
    if end_frame_number is None:
        end_frame_number = len(dataset)
    for frame in range(start_frame_number, end_frame_number):
        pyntcloud = dataset[frame].to_instance("PYNTCLOUD")
        orig_file = Path(ros_file_path).stem
        filename = folder_to_write_path.joinpath(
            f"{orig_file}_{frame}.{output_format.lower()}"
        )
        console.print(
            f"frame {frame} of {Path(ros_file_path).name} converted to {filename}"
        )
        pyntcloud.to_file(filename.as_posix())


typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    app()
    typer_click_object()



"""
Functions to calculate distances of points in pointcloud and a plane.
"""

import numpy as np

from pointcloudset.geometry import plane


def calculate_distance_to_plane(
    pointcloud,
    target: np.ndarray,
    absolute_values: bool = True,
    normal_dist: bool = True,
):
    """Calculate the distance of each point to a plane.

    Note:
        Adds the result as a new column to the data of the pointcloud. Uses the plane
        equation a x + b y + c z + d = 0.

    Args:
        pointcloud (PointCloud): PointCloud for which the distance to the plane is calculated.
        target (numpy.ndarray): [a, b, c, d] as parameters for the target plane, could be
            provided by plane_segmentation.
        absolute_values (bool, optional): Calculate absolute distances if ``True``.
            Defaults to ``True``.
        normal_dist (bool, optional): Calculate normal distance if ``True``, calculate
            distance in direction of line of sight if ``False``. Defaults to ``True``.

    Returns:
        PointCloud: PointCloud including distances to a plane for each point.
    """
    points = pointcloud.points.xyz
    distances = np.asarray(
        [plane.distance_to_point(point, target, normal_dist) for point in points]
    )
    if absolute_values:
        distances = np.absolute(distances)
    plane_str = np.array2string(target, formatter={"float_kind": lambda x: "%.4f" % x})
    plane_str = " ".join(plane_str.split())  # delete multiple white space
    pointcloud._add_column(f"distance to plane: {plane_str}", distances)
    return pointcloud



"""
Functions for file input and output.
"""

from pyntcloud.io import (
    FROM_FILE as POINTCLOUD_FROM_FILE,
)  # needs to be here, ignore warnings

from pointcloudset.io.dataset.dir import dataset_from_dir, dataset_to_dir
from pointcloudset.io.dataset.pointcloud import dataset_from_pointclouds
from pointcloudset.io.dataset.ros import dataset_from_ros
from pointcloudset.io.pointcloud.csv import write_csv
from pointcloudset.io.pointcloud.open3d import from_open3d, to_open3d
from pointcloudset.io.pointcloud.pandas import from_dataframe, to_dataframe
from pointcloudset.io.pointcloud.pyntcloud import from_pyntcloud, to_pyntcloud

DATASET_FROM_FILE = {
    "BAG": dataset_from_ros,
    "DIR": dataset_from_dir,
    "ROS2": dataset_from_ros,
}

DATASET_TO_FILE = {"DIR": dataset_to_dir}

DATASET_FROM_INSTANCE = {"POINTCLOUDS": dataset_from_pointclouds}

POINTCLOUD_TO_FILE = {"CSV": write_csv}

POINTCLOUD_FROM_INSTANCE = {
    "PYNTCLOUD": from_pyntcloud,
    "OPEN3D": from_open3d,
    "DATAFRAME": from_dataframe,
    "PANDAS": from_dataframe,
}
POINTCLOUD_TO_INSTANCE = {
    "PYNTCLOUD": to_pyntcloud,
    "OPEN3D": to_open3d,
    "DATAFRAME": to_dataframe,
    "PANDAS": to_dataframe,
}



from __future__ import annotations

from typing import TYPE_CHECKING

import pyntcloud

if TYPE_CHECKING:
    from pointcloudset import PointCloud


def from_pyntcloud(pyntcloud_data: pyntcloud.PyntCloud) -> dict:
    """Converts a pyntcloud pointcloud to a PointCloud.

    Args:
        pyntcloud_data (pyntcloud.PyntCloud): Pyntcloud pointcloud which should be converted.

    Returns:
        dict: Pointcloud data.
    """
    if not isinstance(pyntcloud_data, pyntcloud.PyntCloud):
        raise TypeError(
            f"Type {type(pyntcloud_data)} not supported for conversion."
            f"Expected pyntcloud.PyntCloud"
        )
    return {"data": pyntcloud_data.points}


def to_pyntcloud(pointcloud: PointCloud) -> pyntcloud.PyntCloud:
    """Converts a PointCloud to a pyntcloud pointcloud.

    Args:
        pointcloud (PointCloud): PointCloud which should be converted.

    Returns:
        pyntcloud.PointCloud: Pyntcloud pointcloud.
    """
    return pointcloud.points



"""
Utility functions for geometric calculations with planes.
"""

import math

import numpy as np

from pointcloudset.geometry import point


def distance_to_point(
    point_a: np.ndarray, plane_model: np.ndarray, normal_dist: bool = True
) -> float:
    """Calculate the distance from a plane to a point.
    https://mathworld.wolfram.com/Point-PlaneDistance.html

    Note:
        Uses the plane equation a x + b y + c z + d = 0.

    Args:
        point_a (numpy.ndarray): [x, y, z], point for which the distance is calculated
            to plane.
        plane_model (numpy.ndarray):  [a, b, c, d] parameters of the plane equation,
            could be provided by :func:`pointcloudset.pointcloud.PointCloud.plane_segmentation`.
        normal_dist (bool): Calculate normal distance if ``True``, calculate
            distance in direction of line of sight if ``False``. Defaults to ``True``.

    Returns:
        float: Distance between plane and point.

    Raises:
        ValueError: If point does not have 3 values or if plane does not have 4 values.
    """
    if len(point_a) != 3:
        raise ValueError("point needs to have 3 values")
    if len(plane_model) != 4:
        raise ValueError("plane_model needs to have 4 values")
    if normal_dist:
        distance = (
            plane_model[0] * point_a[0]
            + plane_model[1] * point_a[1]
            + plane_model[2] * point_a[2]
            + plane_model[3]
        ) / (math.sqrt(plane_model[0] ** 2 + plane_model[1] ** 2 + plane_model[2] ** 2))
    else:
        line_of_sight = point_a
        intersection_point = intersect_line_of_sight(line_of_sight, plane_model)
        distance = point.distance_to_point(intersection_point, point_a)
    return distance


def intersect_line_of_sight(line: np.ndarray, plane_model: np.ndarray) -> np.ndarray:
    """Calculate the point of intersection between a line and a plane.

    Note:
        Uses the plane equation a x + b y + c z + d = 0.

    Args:
        line (numpy.ndarray): [lx, ly, lz], line of sight through origin and point
            (lx,ly,lz).
        plane_model (numpy.ndarray): [a, b, c, d] parameters of the plane equation,
            could be provided by :func:`pointcloudset.pointcloud.PointCloud.plane_segmentation`.

    Returns:
        numpy.ndarray: [px, py, pz], point of intersection of line of sight and plane.

    Raises:
        ValueError: If line does not have 3 values or if plane does not have 4 values.
    """
    if len(line) != 3:
        raise ValueError("line needs to have 3 values")
    if len(plane_model) != 4:
        raise ValueError("plane_model needs to have 4 values")
    t = (-plane_model[3]) / (
        line[0] * plane_model[0] + line[1] * plane_model[1] + line[2] * plane_model[2]
    )
    px = line[0] * t
    py = line[1] * t
    pz = line[2] * t

    return np.array([px, py, pz])



import json
import uuid
import warnings
from datetime import datetime
from pathlib import Path

import dask.dataframe as dd
import numpy as np
import pandas as pd
from rich import print

import pointcloudset

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
DELIMITER = ";"


def dataset_to_dir(
    dataset_in, file_path: Path, use_orig_filename: bool = True, **kwargs
) -> Path:
    """Writes Dataset to directory.

    Args:
        dataset_in (Dataset): Dataset to write.
        file_path (pathlib.Path): Destination path.
        use_orig_filename (bool): Use filename from which the dataset was read. Defaults to ``True``.
        **kwargs: Keyword arguments to pass to dask to_parquet function
    """
    if not dataset_in.has_pointclouds():
        raise ValueError("dataset must have data ")
    _check_dir(file_path)
    orig_filename = Path(dataset_in.meta["orig_file"]).stem
    if len(orig_filename) == 0:
        orig_filename = str(uuid.uuid4())
    folder = file_path.joinpath(orig_filename) if use_orig_filename else file_path
    empty_data = _get_empty_data(dataset_in)
    dataset_to_write = dataset_in._replace_empty_frames_with_nan(empty_data)
    data = dd.from_delayed(dataset_to_write.data)
    data.to_parquet(folder, **kwargs)
    meta = dataset_in.meta
    meta["timestamps"] = [
        timestamp.strftime(DATETIME_FORMAT) for timestamp in dataset_in.timestamps
    ]
    meta["empty_data"] = empty_data.to_dict()
    meta["version"] = pointcloudset.__version__
    with open(folder.joinpath("meta.json"), "w") as outfile:
        json.dump(dataset_in.meta, outfile)
    _check_dir_contents(folder)
    return folder.parent


def dataset_from_dir(dir: Path, ext: str) -> dict:
    # sourcery skip: simplify-len-comparison
    """Reads a Dataset from a directory.

    Args:
        dir (pathlib.Path): Path of directory.

    Returns:
        dict: Lidar data with timestamps and metadata.
    """
    _check_dir(dir)
    dirs = [e for e in dir.iterdir() if e.is_dir()]

    if len(dirs) > 0:
        dirs.sort(key=_get_folder_number)
    else:
        dirs = [dir]

    data = []
    timestamps = []
    meta = []
    for path in dirs:
        res = _dataset_from_single_dir(path)
        data.extend(res["data"])
        timestamps.extend(res["timestamps"])
        meta.append(res["meta"])
    meta = meta[0]
    del meta["timestamps"]
    if "empty_data" in meta:
        empty_data = pd.DataFrame.from_dict(meta["empty_data"])
    else:
        empty_data = pd.DataFrame()
        # for backwards compatibility
    return {
        "data": data,
        "timestamps": timestamps,
        "meta": meta,
        "empty_data": empty_data,
    }


def _get_folder_number(path: Path) -> int:
    try:
        return int(path.stem)
    except ValueError:
        raise ValueError(f"{path} is not a path with a dataset")


def _dataset_from_single_dir(dir: Path) -> dict:
    _check_dir(dir)
    parquet_files = list(dir.glob("*.parquet"))
    data = dd.read_parquet(parquet_files)
    with open(dir.joinpath("meta.json"), "r") as infile:
        meta = json.loads(infile.read())
    timestamps_raw = meta["timestamps"]
    timestamps = [
        datetime.strptime(timestamp, DATETIME_FORMAT) for timestamp in timestamps_raw
    ]
    return {
        "data": data.to_delayed(),
        "timestamps": timestamps,
        "meta": meta,
    }


def _check_dir(file_path: Path):
    if not isinstance(file_path, Path):
        raise TypeError("expecting a pathlib Path object")
    if len(file_path.suffix) != 0:
        raise ValueError("expecting a path not a filename")


def _check_dir_contents(dir: Path):
    """Quick test if all"""
    sub_dirs = list(dir.glob("**"))
    for path in sub_dirs:
        if len(sub_dirs) != 1 and path != dir or len(sub_dirs) == 1:
            _check_dir_contents_single(path)


def _check_dir_contents_single(dir: Path):
    """checking the folder content of a written dataset."""
    assert dir.joinpath("meta.json").is_file(), f"meta.json is missing in {dir}"
    assert dir.joinpath("part.0.parquet"), f"part.0.parquet is missing in {dir}"


def _get_empty_data(dataset_in) -> pd.DataFrame:
    """Get an row of data to use as placeholder for empty datasets.

    Args:
        dataset_in (Dataset): Dataset which contains at least on valid pointcloud.

    Returns:
        pd.DataFrame: empta data placeholder
    """
    empty_data = pd.DataFrame.from_dict(
        {
            "x": [np.nan],
            "y": [np.nan],
            "z": [np.nan],
            "original_id": [np.nan],
        }
    )
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        for pointcloud in dataset_in:
            if len(pointcloud) > 0:
                empty_data = pd.DataFrame(pointcloud.data.iloc[0]).T
                break
    return empty_data



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal, get_type_hints

import numpy as np
import pandas
import plotly.graph_objects as go
from dask import delayed

from pointcloudset.dataset_core import DatasetCore
from pointcloudset.io import DATASET_FROM_FILE, DATASET_FROM_INSTANCE, DATASET_TO_FILE
from pointcloudset.pipeline.delayed_result import DelayedResult
from pointcloudset.plot.dataset import animate_dataset
from pointcloudset.pointcloud import PointCloud


def _is_pipline_returing_pointcloud(pipeline, warn=True) -> bool:
    type_hints = get_type_hints(pipeline)
    res = False
    if "return" in type_hints:
        res = get_type_hints(pipeline)["return"] == PointCloud
    elif warn:
        print(
            (
                f"No return type was defined in {pipeline.__name__}:"
                "will not return a new dataset"
            )
        )
    return res


class Dataset(DatasetCore):
    """
    Dataset Class which contains multiple pointclouds, timestamps and metadata.
    For more details on how to use the Dataset Class please refer to the usage.ipynb
    notebook for an interactive tutorial. The notebook can also be found in the tutorial
    section of the docu.
    """

    def __getitem__(self, pointcloud_number: slice | int) -> DatasetCore | PointCloud:
        if isinstance(pointcloud_number, slice):
            data = self.data[pointcloud_number]
            timestamps = self.timestamps[pointcloud_number]
            meta = self.meta
            return Dataset(data, timestamps, meta)
        elif isinstance(pointcloud_number, int):
            df = self.data[pointcloud_number].compute()
            timestamp = self.timestamps[pointcloud_number]
            return PointCloud(
                data=df, orig_file=self.meta["orig_file"], timestamp=timestamp
            )
        else:
            raise TypeError(f"Wrong type {type(pointcloud_number).__name__}")

    @classmethod
    def from_file(cls, file_path: Path, **kwargs):
        """Reads a Dataset from a file.gfile
        For larger ROS bagfiles files use the commandline tool pointcloudset to convert
        the ROS file beforehand.

        Supported are the native format which is a directore filled with fastparquet frames and
        ROS bag files (.bag).


        Args:
            file_path (pathlib.Path): File path where Dataset should be read from.\n
                If file format is a directory: :func:`pointcloudset.io.dataset.dir.dataset_from_dir`\n
                If file format is a ROS bag file: :func:`pointcloudset.io.dataset.bag.dataset_from_rosbag`
            **kwargs: Keyword arguments to pass to func.

        Returns:
            Dataset: Dataset object from file.

        Raises:
            ValueError: If file format is not supported.
            TypeError: If file_path is not a Path object.

        Examples:

            .. code-block:: python

                pointcloudset.Dataset.from_file(bag_file, topic="lidar/points", keep_zeros=False)

        Examples:

            .. code-block:: python

                pointcloudset.Dataset.from_file(bag_file, topic="lidar/points", keep_zeros=False)
        """
        from_dir = False
        from_dir = False
        if not isinstance(file_path, Path):
            raise TypeError("Expecting a Path object for file_path")
        ext = file_path.suffix[1:].upper()
        if ext == "":
            if file_path.joinpath("meta.json").exists():
                from_dir = True
                ext = "DIR"  # native pointcloudset format
            else:
                ext = "ROS2"  # ROS2 is also a directory for both mcap and dp3
                if not file_path.joinpath("metadata.yaml").exists():
                    raise FileNotFoundError(
                        "metadata.yaml not found in directory, which is required for ROS2 format"
                    )
        if ext not in DATASET_FROM_FILE:
            raise ValueError(
                (
                    f"Unsupported file format {ext}; supported formats are:"
                    " {DATASET_FROM_FILE.keys()}"
                )
            )
        res = DATASET_FROM_FILE[ext](file_path, ext=ext, **kwargs)
        meta = res["meta"]
        out = cls(data=res["data"], timestamps=res["timestamps"], meta=meta)
        if from_dir:
            out = out._replace_nan_frames_with_empty(res["empty_data"])
        return out

    def to_file(self, file_path: Path = Path(), **kwargs) -> None:
        """Writes a Dataset to a file.

        Supported is the native format which is a directory full of fastparquet files
        with meta data.

        Args:
            file_path (pathlib.Path): File path where Dataset should be saved.\n
                If file format is a directory: :func:`pointcloudset.io.dataset.dir.dataset_to_dir`
            **kwargs: Keyword arguments to pass to func.
        """
        DATASET_TO_FILE["DIR"](self, file_path=file_path, **kwargs)

    @classmethod
    def from_instance(
        cls,
        library: str,
        instance: list[PointCloud],
        **kwargs,
    ) -> Dataset:
        """Converts a library instance to a pointcloudset Dataset.

        Args:
            library (str): Name of the library.\n
                If "pointclouds": :func:`pointcloudset.io.dataset.pointclouds.dataset_from_pointclouds`
            instance (list[PointCloud]): Instance from which to convert.
            **kwargs: Keyword arguments to pass to func.

        Returns:
            Dataset: Dataset object derived from the instance.

        Raises:
            ValueError: If instance is not supported.

        Examples:

            .. code-block:: python

                pointcloudset.Dataset.from_instance("pointclouds", [pc1, pc2])

        """
        library = library.upper()
        if library not in DATASET_FROM_INSTANCE:
            raise ValueError(
                "Unsupported library; supported libraries are: {}".format(
                    list(DATASET_FROM_INSTANCE)
                )
            )
        else:
            return cls(**DATASET_FROM_INSTANCE[library](instance, **kwargs))

    def apply(
        self,
        func: Callable[[PointCloud], PointCloud] | Callable[[PointCloud], Any],
        warn: bool = True,
        **kwargs,
    ) -> Dataset | DelayedResult:
        """Applies a function to the dataset. It is also possible to pass keyword
        arguments.

        Args:
            func (Union[Callable[[PointCloud], PointCloud], Callable[[PointCloud], Any]]): Function to
                apply. If it returns a PointCloud and has the according type hint a new
                Dataset will be generated.
            warn (bool): If ``True`` warning if result is not a Dataset, if ``False``
                warning is turned off.
            **kwargs: Keyword arguments to pass to func.

        Returns:
            Union[Dataset, DelayedResult]: A Dataset if the function returns a PointCloud,
            otherwise a DelayedResult object which is a tuple of dask delayed objects.

        Examples:

            .. code-block:: python

                def func(pointcloud:pointcloudset .PointCloud) -> pointcloudset.PointCloud:
                    return pointcloud.limit(x,0,1)

                dataset.apply(func)
                # This results in a new Dataset

            .. code-block:: python

                def func(pointcloud:pointcloudset.PointCloud) -> float:
                    return pointcloud.data.x.max()

                dataset.apply(func)

            .. code-block:: python

                def func(pointcloud:pointcloudset.PointCloud, test: float) -> float:
                    return pointcloud.data.x.max() + test

                dataset.apply(func, test=10)
        """
        returns_pointcloud = _is_pipline_returing_pointcloud(func, warn=warn)
        columns = list(self[0].data.columns)

        if returns_pointcloud:

            def pipeline_delayed(element_in, timestamp):
                pointcloud_in = PointCloud(data=element_in, timestamp=timestamp)
                pointcloud = func(pointcloud_in, **kwargs)
                if not pointcloud._has_data():
                    pointcloud = PointCloud(columns=columns)
                return pointcloud.data  # to generate an empty pointcloud

        else:

            def pipeline_delayed(element_in, timestamp):
                pointcloud = PointCloud(data=element_in, timestamp=timestamp)
                return func(pointcloud, **kwargs)

        res = []
        for i in range(0, len(self)):
            item = delayed(pipeline_delayed)(self.data[i], self.timestamps[i])
            res.append(item)

        if returns_pointcloud:
            return Dataset(data=res, timestamps=self.timestamps, meta=self.meta)
        else:
            return DelayedResult(res)

    @property
    def has_original_id(self) -> bool:
        """Check if all pointclouds in the Dataset have original_ids

        Returns:
            bool: ``True`` if all PointClouds in the the Dataset returns has_original_id.
        """

        def check_original_id(pc):
            return pc.has_original_id

        return all(self.apply(check_original_id, warn=False).compute())

    def agg(
        self,
        agg: str | list | dict,
        depth: Literal["dataset", "pointcloud", "point"] = "dataset",
    ) -> Union[
        pandas.Series, list[pandas.DataFrame], pandas.DataFrame, pandas.DataFrame
    ]:
        """Aggregate using one or more operations over the whole dataset.
        Similar to :meth:`pandas.DataFrame.aggregate`.
        Uses :class:`dask.dataframe.DataFrame` with parallel processing.


        Args:
            agg (Union[str, list, dict]): Function to use for aggregating.
            depth (Literal["dataset", "pointcloud", "point"], optional): Aggregation level: "dataset", "pointcloud" or
                "point". Defaults to "dataset".

        Returns:
            Union[pandas.DataFrame, pandas.DataFrame, pandas.Series]: Results of the
            aggregation. This can be a pandas DataFrame or Series, depending on the
            depth and aggregation.

        Raises:
            ValueError: If depth is not "dataset", "pointcloud" or "point".

        Examples:

            .. code-block:: python

                dataset.agg("max", "pointcloud")

            .. code-block:: python

                dataset.agg(["min","max","mean","std"])

            .. code-block:: python

                dataset.agg({"x" : ["min","max","mean","std"]})
        """
        if depth == "point":
            if self.has_original_id:
                data = self._agg(agg).compute()
                if isinstance(agg, str):
                    data.columns = [
                        i if i in ["N", "original_id"] else f"{i} {agg}"
                        for i in data.columns
                    ]

                return data
            else:
                raise ValueError("this operations nees original_id in each pointcloud")
        elif depth == "pointcloud":
            return self._agg_per_pointcloud(agg)
        elif depth == "dataset":
            data = self._agg(agg).compute()

            if isinstance(agg, dict):
                data = data.iloc[:, :-2]
                if isinstance(data, pandas.DataFrame):
                    data.columns = data.columns.to_flat_index()
                agg_list = list(agg.values())[0]
                data = data.agg(agg_list)
            else:
                data = data.agg(agg)

            if not isinstance(agg, list):
                data.index = [f"{i} {agg}" for i in data.index]
            return data
        else:
            raise ValueError("depth needs to be dataset, pointcloud or point")

    def min(self, depth: str = "dataset"):
        """Aggregate using min operation over the whole dataset.
        Similar to :meth:`pandas.DataFrame.aggregate`.
        Uses :class:`dask.dataframe.DataFrame` with parallel processing.

        Args:
            depth (Literal["dataset", "pointcloud", "point"], optional): Aggregation level:
            "dataset", "pointcloud" or "point". Defaults to "dataset".

        Returns:
            Union[pandas.DataFrame, pandas.DataFrame, pandas.Series]: Aggregated
            Dataset.

        Examples:

            .. code-block:: python

                dataset.min()

            .. code-block:: python

                dataset.min("pointcloud")

            .. code-block:: python

                dataset.min("point")

        Hint:

            See also: :func:`pointcloudset.dataset.Dataset.agg`\n
            Same as:

            .. code-block:: python

                dataset.agg(["min"])
        """
        return self.agg("min", depth=depth)

    def max(self, depth: str = "dataset"):
        """Aggregate using max operation over the whole dataset.
        Similar to :meth:`pandas.DataFrame.aggregate`.
        Uses :class:`dask.dataframe.DataFrame` with parallel processing.

        Args:
            depth (Literal["dataset", "pointcloud", "point"], optional): Aggregation level:
            "dataset", "pointcloud" or "point". Defaults to "dataset".

        Returns:
            Union[pandas.DataFrame, pandas.DataFrame, pandas.Series]: Aggregated
            Dataset.

        Examples:

            .. code-block:: python

                dataset.max()

            .. code-block:: python

                dataset.max("pointcloud")

            .. code-block:: python

                dataset.max("point")

        Hint:

            See also: :func:`pointcloudset.dataset.Dataset.agg`\n
            Same as:

            .. code-block:: python

                dataset.agg(["max"])
        """
        return self.agg("max", depth=depth)

    def mean(self, depth: str = "dataset"):
        """Aggregate using mean operation over the whole dataset.
        Similar to :meth:`pandas.DataFrame.aggregate`.
        Uses :class:`dask.dataframe.DataFrame` with parallel processing.

        Args:
            depth (Literal["dataset", "pointcloud", "point"], optional): Aggregation level:
            "dataset", "pointcloud" or "point". Defaults to "dataset".

        Returns:
            Union[pandas.DataFrame, pandas.DataFrame, pandas.Series]: Aggregated Dataset.

        Examples:

            .. code-block:: python

                dataset.mean()

            .. code-block:: python

                dataset.mean("pointcloud")

            .. code-block:: python

                dataset.mean("point")

        Hint:

            See also: :func:`pointcloudset.dataset.Dataset.agg`\n
            Same as:

            .. code-block:: python

                dataset.agg(["mean"])
        """
        return self.agg("mean", depth=depth)

    def std(self, depth: str = "dataset"):
        """Aggregate using std operation over the whole dataset.
        Similar to :meth:`pandas.DataFrame.aggregate`.
        Uses :class:`dask.dataframe.DataFrame` with parallel processing.

        Args:
            depth (Literal["dataset", "pointcloud", "point"], optional): Aggregation level:
            "dataset", "pointcloud" or "point". Defaults to "dataset".

        Returns:
            Union[pandas.DataFrame, pandas.DataFrame, pandas.Series]: Aggregated Dataset.

        Examples:

            .. code-block:: python

                dataset.std()

            .. code-block:: python

                dataset.std("pointcloud")

            .. code-block:: python

                dataset.std("point")

        Hint:

            See also: :func:`pointcloudset.dataset.Dataset.agg`\n
            Same as:

            .. code-block:: python

                dataset.agg(["std"])
        """
        return self.agg("std", depth=depth)

    def _agg_per_pointcloud(
        self, agg: str | list | dict
    ) -> pandas.DataFrame | list | pandas.DataFrame:
        def get(pointcloud, agg: str | list | dict):
            return pointcloud.data.agg(agg)

        res = self.apply(get, warn=False, agg=agg).compute()
        if isinstance(agg, str):
            res = pandas.DataFrame(res)
            if not isinstance(agg, dict) and "original_id" in res.columns:
                res = res.drop("original_id", axis=1)
            res.columns = [f"{column} {agg}" for column in res.columns]
            res.index.name = "pointcloud"
            res["timestamp"] = self.timestamps
        return res

    def extend(self, dataset: Dataset) -> Dataset:
        """Extends the dataset by another one.

        Args:
            dataset (Dataset): Dataset to extend another dataset.

        Returns:
            Dataset: Extended dataset.
        """
        key = "extended"
        meta = self.meta
        if key in meta:
            new = meta[key]
            new.extend(dataset.meta)
            meta[key] = new
        else:
            meta[key] = [dataset.meta]
        self.data.extend(dataset.data)
        self.timestamps.extend(dataset.timestamps)
        self._check()
        return self

    def animate(self, **kwargs) -> go.Figure:
        """Plot and animate a PointClouds in a dataset as a 3D scatter plot with
        `Plotly <https://plotly.com/>`_.
        It uses the plot function of PointCloud and bundles them together for an
        interactive animation :func:`pointcloudset.pointcloud.plot`.

        You can also pass arguments to the `Plotly <https://plotly.com/>`_
        express function :func:`plotly.express.scatter_3d`.

        Args:
            **kwargs: Keyword arguments to pass to plot of a single pointcloud and
                plotly express.

        Returns:
            plotly.graph_objs.Figure: The interactive Plotly plot, best used inside a
            Jupyter Notebook.

        Returns:
            go.Figure: _description_

        Examples:

            .. code-block:: python

                dataset_bag.animate(hover_data=True, color="intensity")
        """
        return animate_dataset(self, **kwargs)

    def _replace_empty_frames_with_nan(self, empty_data: pandas.DataFrame):
        """Function to replace empty pointclouds with pointclouds wiht 1 point with all
        nan values. Needed to save files with dask.
        """

        def _exchange_empty_pointclouds_with_nan(frame: PointCloud) -> PointCloud:
            if not frame._has_data():
                frame.data = empty_data
            return frame

        return self.apply(_exchange_empty_pointclouds_with_nan)

    def _replace_nan_frames_with_empty(self, empty_data: pandas.DataFrame):
        """Function to replace nan pointclouds with empty pointcouds
        Needed to after reading dataset files.
        """

        def _exchange_nan_pointclouds_with_empty(frame: PointCloud) -> PointCloud:
            if (len(frame) == 1) and np.allclose(empty_data.values, frame.data.values):
                frame = PointCloud(columns=frame.data.columns)
            return frame

        return self.apply(_exchange_nan_pointclouds_with_empty)



"""
Functions to calculate differences between frames.
"""

import numpy as np
import pandas as pd


def calculate_distance_to_pointcloud(pointcloud, target):
    """Calculate the differences for each point in a pointcloud which also exists in
    the target pointcloud (pointcloud - target). Points with the same original_id are compared.

    Note:
        Adds the results to the data of the pointcloud.

    Args:
        pointcloud (PointCloud): PointCloud for which the differences to the target are calculated.
        target (PointCloud): PointCloud which is substracted from pointcloud.

    Returns:
        PointCloud: PointCloud including differences to target.

    Raises:
        ValueError: If there are no original_ids in pointcloud or in target or if they have
            no common original_ids.
        NotImplementedError: If difference has already been calculated.
    """
    if "x difference" in pointcloud.data.columns:
        raise NotImplementedError("Differences of differences are not supported.")
    if not pointcloud.has_original_id:
        raise ValueError("PointCloud does not contain original_id.")
    if not target.has_original_id:
        raise ValueError("Target does not contain original_id.")
    reference_original_ids = pointcloud.data.original_id.values
    target_original_ids = target.data.original_id.values
    intersection = np.intersect1d(reference_original_ids, target_original_ids)
    if len(intersection) > 0:
        return _calculate_difference(intersection, pointcloud, target)

    else:
        raise ValueError("No common original_ids in pointcloud and target.")


def _calculate_difference(intersection: np.ndarray, pointcloud, target):
    diff_list = [
        _calculate_single_point_difference(pointcloud, target, id)
        for id in intersection
    ]
    original_types = [str(types) for types in diff_list[0].dtypes.values]
    target_type_dict = dict(zip(diff_list[0].columns.values, original_types))
    diff_df = pd.concat(diff_list)
    diff_df = diff_df.astype(target_type_dict)
    diff_df = diff_df.reset_index(drop=True)
    pointcloud.data = pointcloud.data.merge(diff_df, on="original_id", how="left")
    return pointcloud


def _calculate_single_point_difference(
    pointcloud, pointcloud_b, original_id: int
) -> pd.DataFrame:
    """Calculate the difference of one element of a Point in the current PointCloud to
    the corresponding point in PointCloud B. Both frames must contain the same original_id.

    Args:
        pointcloud_b (PointCloud): PointCloud which contains the point to comapare to.
        original_id (int): Original ID of the point.

    Returns:
        pd.DataFrame: A single row DataFrame with the differences (A - B).
    """
    point_a = pointcloud.extract_point(original_id, use_original_id=True)
    try:
        point_b = pointcloud_b.extract_point(original_id, use_original_id=True)
        difference = point_a - point_b
    except IndexError:
        # there is no point with the orignal_id in frameB
        difference = point_a
        difference.loc[:] = np.nan
    difference = difference.drop(["original_id"], axis=1)
    difference.columns = [f"{column} difference" for column in difference.columns]
    difference["original_id"] = point_a["original_id"]
    return difference



"""
Functions for plotting datasets.
"""
import warnings

import plotly.graph_objs as go


def animate_dataset(dataset, **kwargs) -> go.Figure:
    """Geneartes animation used in the Dataset class.

    Args:
        dataset (Dataset): dataset to animate

    Returns:
        go.Figure: interactive plotly plot
    """

    warnings.warn("Experimental Feature")

    def plot_frame(pointcloud):
        return pointcloud.plot(**kwargs)

    start_frame = 0
    frames = dataset.apply(plot_frame, warn=False).compute()

    fig = go.Figure()

    for frame in frames:
        fig.add_trace(frame["data"][0])
    fig.data[start_frame].visible = True

    fig.update_layout(sliders=_gen_sliders(dataset, fig))

    return _limit_to_bounding_box(dataset, fig)


def _limit_to_bounding_box(dataset, fig: go.Figure) -> go.Figure:
    """limit the view to the bounding box of the complete dataset."""
    bounding_box = dataset.bounding_box
    fig.update_layout(
        title=_gen_title(dataset, 0),
        scene=dict(
            xaxis=dict(
                range=list(bounding_box["x"].values),
            ),
            yaxis=dict(
                range=list(bounding_box["y"].values),
            ),
            zaxis=dict(
                range=list(bounding_box["z"].values),
            ),
        ),
    )
    return fig


def _gen_sliders(dataset, fig):
    """Generate sliders for plotly."""
    steps = []
    for i in range(len(fig.data)):
        step = dict(
            method="update",
            label=str(i),
            args=[
                {"visible": [False] * len(fig.data)},
                {"title": _gen_title(dataset, i)},
            ],  # layout attribute
        )
        step["args"][0]["visible"][i] = True  # Toggle i'th trace to "visible"
        steps.append(step)

    sliders = [
        dict(
            active=0,
            currentvalue={"prefix": "Frame: "},
            pad={"t": 50},
            steps=steps,
        )
    ]
    return sliders


def _gen_title(dataset, frame_number: int) -> str:
    """Title string of frames."""
    return f"Frame: {frame_number} {dataset[frame_number].timestamp_str}"



import codecs
import os

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setup(
    name="pointcloudset",
    packages=find_packages(include=["pointcloudset", "pointcloudset*"]),
    version=get_version("pointcloudset/__init__.py"),
    author="VIRTUAL VEHICLE Research GmbH",
    author_email="thomas.goelles@v2c2.at",
    description="Analyze large datasets of point clouds recorded over time in an efficient way",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    project_urls={
        "Bug Tracker": "https://github.com/virtual-vehicle/pointcloudset/issues",
        "Source": "https://github.com/virtual-vehicle/pointcloudset",
        "Documentation": "https://virtual-vehicle.github.io/pointcloudset/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    python_requires=">=3.9",
    install_requires=[
        "numpy",
        "pandas",
        "pyntcloud",
        "open3d>=0.15.0",
        "plotly",
        "fastparquet",
        "dask>=2022.02.0",
        "typer",
        "rich",
        "rosbags",
    ],
    extras_require={"LAS": ["pylas"]},
    entry_points={
        "console_scripts": ["pointcloudset = pointcloudset.io.dataset.commandline:app"]
    },
)



"""
Functions to calculate distances between points in pointcloud and a point.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import pointcloudset


def calculate_distance_to_point(
    pointcloud: pointcloudset.pointcloud.PointCloud, target: np.ndarray, **kwargs
) -> pointcloudset.pointcloud.PointCloud:
    """Calculate the Euclidian distance to a point for each point in the pointcloud.

    Note:
        Adds the result as a new column to the data of the pointcloud.

    Args:
        pointcloud (PointCloud): PointCloud for which the Euclidean distance to the point is
            calculated.
        target (numpy.ndarray): [x, y, z] as coordinates of the point to which the
            Euclidean distance is calculated.

    Returns:
        PointCloud: PointCloud including Euclidean distances to a point for each point.
    """
    point_a = target
    points = pointcloud.points.xyz
    distances = np.array([np.linalg.norm(point_a - point) for point in points])
    point_str = np.array2string(target, formatter={"float_kind": lambda x: "%.4f" % x})
    point_str = " ".join(point_str.split())  # delete multiple white space
    pointcloud._add_column(f"distance to point: {point_str}", distances)
    return pointcloud



from __future__ import annotations

from collections import UserList
from typing import Any

import dask


class DelayedResult(UserList):
    def __init__(self, data):
        self.data = data

    def compute(self) -> list:
        return list(dask.compute(*self.data))

    def __getitem__(self, pointcloud_number: slice | int) -> DelayedResult | Any:
        if isinstance(pointcloud_number, (slice, int)):
            return super().__getitem__(pointcloud_number).compute()
        else:
            raise TypeError(f"Wrong type {type(pointcloud_number).__name__}")



from __future__ import annotations

import traceback
import warnings
from datetime import datetime

import numpy as np
import pandas as pd
import pyntcloud


class PointCloudCore:
    """
    PointCloudCore Class with all the main methods and properties of the
    PointCloud Class.
    """

    def __init__(
        self,
        data: pd.DataFrame = None,
        orig_file: str = "",
        timestamp: datetime = None,
        columns: list = ["x", "y", "z"],
    ):
        self.timestamp = datetime.now() if timestamp is None else timestamp
        """Timestamp."""
        self.orig_file = orig_file
        """Path to orginal file. Defaults to empty."""

        if data is None:
            # "empty" PointCloud with one line of all nans. This is necessary in order
            # to save datasets with dask see issue#6
            values = np.repeat(np.nan, len(columns))
            empty_data = pd.DataFrame([values], columns=columns)
            self.data = empty_data
            self.data = self.data.drop([0])
        else:
            self.data = data
        """The data as a pandas DataFrame."""

        with warnings.catch_warnings():
            # ignore warnings produced by pyntcloud when the pointcloud is empty
            warnings.simplefilter("ignore")
            self.points = pyntcloud.PyntCloud(self.data, mesh=None)
        """PyntCloud object with x,y,z coordinates."""
        self._check_index()

    @property
    def timestamp_str(self) -> str:
        """Timestamp to human readable date and time string.

        Returns:
            str: Date/time string.
        """
        return self.timestamp.strftime("%A, %B %d, %Y %H:%M:%S.%f")

    @property
    def data(self):
        """All the data, x,y,z and auxiliary data such as intensity, range and more."""
        return self.__data

    @data.setter
    def data(self, df: pd.DataFrame):
        if not isinstance(df, pd.DataFrame):
            raise TypeError("Data argument must be a DataFrame")
        elif not {"x", "y", "z"}.issubset(df.columns):
            raise ValueError("Data must have x, y and z coordinates")
        self._update_data(df)
        self._check_index()

    @property
    def bounding_box(self) -> pd.DataFrame:
        """The axis aligned boundary box as a :class:`pandas.DataFrame`."""
        return self.data[["x", "y", "z"]].agg(["min", "max"])

    @property
    def centroid(self) -> np.array:
        """Geometric center for the point cloud."""
        return self.points.centroid

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.data}, "
            f"{self.timestamp}, {self.orig_file})"
        )

    def __str__(self) -> str:
        return (
            f"pointcloud: with {len(self)} points, data:{list(self.data.columns)},"
            f" from {self.timestamp_str}"
        )

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, id: int | slice) -> pd.DataFrame:
        if isinstance(id, slice):
            return self.data.iloc[id]
        elif isinstance(id, int):
            return self.data.iloc[[id]]
        else:
            raise TypeError(f"Wrong type {type(id).__name__}")

    def _update_data(self, df: pd.DataFrame):
        """Utility function. Implicitly called when self.data is assigned."""
        self.__data = df
        self.points = pyntcloud.PyntCloud(self.__data[["x", "y", "z"]], mesh=None)

    def _check_index(self):
        """A private function to check if the index of self.data is sane."""
        if len(self) > 0:
            assert self.data.index[0] == 0, "index should start with 0"
            assert self.data.index[-1] + 1 == len(
                self
            ), "index should be as long as the data"
            assert (
                self.data.index.is_monotonic_increasing
            ), "index should be monotonic increasing"

    def _add_column(self, column_name: str, values: np.array) -> PointCloudCore:
        """Adding a new column with a scalar value to the data of the pointcloud.

        Args:
            column_name (str): name of the new column.
            values (numpy.ndarray): Values of the new column.
        """
        self.data[column_name] = values
        return self

    def _has_data(self) -> bool:
        """Check if pointcloudset pointcloud has data. Data here means point coordinates and
        measurements at each point of the pointcloud.

        Returns:
            bool: ``True`` if the pointcloudset pointcloud contains data.
        """
        return not self.data.empty

    @property
    def has_original_id(self) -> bool:
        """Checks if original_id column is present in the data.
        Original_id identifies a lidar point and makes them comparable.

        Returns:
            bool: ``True`` if the PointCloud contains original_id data, ``False`` if PointCloud
            does not contain original_id data.
        """
        return "original_id" in self.data.columns

    def _contains_original_id_number(self, original_id: int) -> bool:
        """Check if pointcloudset pointcloud contains a specific original_id.

        Args:
            original_id (int): the original_id to check

        Returns:
            bool: ``True`` if the original_id exists.
        """
        return original_id in self.data["original_id"].values

    def describe(self) -> pd.DataFrame:
        """Generate descriptive statistics based on PointCloud.data.describe() and therefore on
        :meth:`pandas:pandas.DataFrame.describe`.

        Returns:
            pandas.DataFrame: Summary statistics of the data of the PointCloud.
        """
        return self.data.describe()

    def extract_point(self, id: int, use_original_id: bool = False) -> pd.DataFrame:
        """Extract a specific point from the PointCloud defined by the point id. The id
        can be the current index of the data from the PointCloud or the original_id.

        Args:
            id (int): ID number
            use_original_id (bool, optional): If ``True`` use original_id, if ``False``
                use current index in PointCloud. Defaults to ``False``.

        Returns:
            pandas.DataFrame: A pointcloud which only contains the defined point.
        """
        try:
            if use_original_id:
                point = self.data[self.data["original_id"] == id]
                if len(point) == 0:
                    raise IndexError
                elif len(point) != 1:
                    raise Exception()
            else:
                point = self[id]
        except IndexError:
            raise IndexError(f"point with {id} does note exist.")
        except Exception:
            print(traceback.print_exc())
        return point.reset_index(drop=True)
